#!/bin/bash

# Copyright © 2021-2024 Devin ApS
# DEVIN UNIX INSTALLER
# INSTALLER VERSION 2.0.0
installer_version="2.0.0"
# Colors :)
RED='\033[0;31m'
GREEN='\033[0;32m'
YELLOW='\033[1;33m'
BLUE='\033[0;34m'
NC='\033[0m' # No Color



# Function to compare version numbers
version_gt() { 
    test "$(echo "$@" | tr " " "\n" | sort -V | head -n 1)" != "$1"; 
}

# Ensure root
ensure_root() {
    if [ "$EUID" -ne 0 ]; then
        color_echo "ERROR" "Please run this script as root."
        exit 1
    fi
}

# Ensure fms is installed by checking which fmsadmin
ensure_fms_installed() {
    if ! which fmsadmin >/dev/null 2>&1; then
    color_echo "ERROR" "FileMaker Server is not installed, please install it first."
    exit 1
    fi
}

# Color echoing function
color_echo() {
    local color=$1
    local msg=$2
    local color_value=""
    local info_box=""

    case $color in
        "ERROR")
            color_value=$RED
            info_box="[-]"
            ;;
        "SUCCESS")
            color_value=$GREEN
            info_box="[+]"
            ;;
        "INFO")
            color_value=$YELLOW
            info_box="[*]"
            ;;
        "QUESTION")
            color_value=$BLUE
            info_box="[?]"
            ;;
        *)
            echo "Invalid color"
            return 1
    esac

    echo -e "${color_value}${info_box} ${msg}${NC}" | tee -a $log_file
}

# Function to check if devin is being ran from a package (full zip download) or as a standalone script
check_packaged_or_standalone() {
    local script_dir="$(dirname "$0")"

    color_echo "INFO" "Detecting if script is ran as standalone or from a package."

    # Check for the presence of the required file and directory
    if [[ -f "$script_dir/bins/devin-engine.fmp12" ]]; then
        # Verify that no other files or directories are present
        color_echo "INFO" "Devin installer is running from a packaged install"
        script_type=package
        if [[ "$OS" == "Darwin" && "$(uname -m)" == "arm64" ]]; then
            script_folder_name="bins/mac_arm/"
        elif [[ "$OS" == "Darwin" && "$(uname -m)" == "x86_64" ]]; then
            script_folder_name="bins/mac_x64/"
        elif [[ "$OS" == "Linux" && "$(uname -m)" == "aarch64" ]]; then
            script_folder_name="bins/ubuntu_arm/"
        elif [[ "$OS" == "Linux" && "$(uname -m)" == "x86_64" ]]; then
            script_folder_name="bins/ubuntu_x64/"
        else
            color_echo "ERROR" "Unsupported OS or architecture."
            exit 1
        fi
    else
        color_echo "INFO" "Devin installer is running as a standalone script"
        script_type=standalone
    fi
}

# Function to uninstall the application
uninstall() {
    color_echo "INFO" "Uninstalling Devin."

    # Handle service differences between macOS and Linux
    if [[ "$OS" == "Darwin" ]]; then
        local PLIST_FILE="/Library/LaunchDaemons/com.devin.handler.plist"

        color_echo "INFO" "Removing Devin Handler service."
        if [ -f "$PLIST_FILE" ]; then
            launchctl unload "$PLIST_FILE" >> $log_file 2>&1
            rm "$PLIST_FILE" >> $log_file 2>&1
            color_echo "SUCCESS" "Devin Handler service has been removed."
        else
            color_echo "INFO" "Devin Handler service plist file does not exist, nothing to remove."
        fi

        if dscl . -read /Groups/fmsadmin GroupMembership | grep -q "devin"; then
            dseditgroup -o edit -d devin -t user fmsadmin >> $log_file 2>&1
            color_echo "INFO" "Removed Devin user from fmsadmin group."
        fi
        rm -rf "$DEVIN_HOME" >> $log_file 2>&1
        sysadminctl -deleteUser devin -secure >> $log_file 2>&1
        
    elif [[ "$OS" == "Linux" ]]; then
        local service_name="devin-handler"

        if systemctl list-unit-files | grep -Fqw "$service_name"; then
            color_echo "INFO" "Stopping and disabling the handler service."
            sudo systemctl disable "$service_name"
            sudo systemctl stop "$service_name"
            local service_file_path=$(systemctl show -p FragmentPath "$service_name" | cut -d= -f2)
            if [ -f "$service_file_path" ]; then
                rm "$service_file_path"
                systemctl daemon-reload
                systemctl reset-failed
            fi
        fi
        userdel -r devin
        groupdel devin
    fi

    # Common steps for both macOS and Linux
    fmsadmin close devin-engine.fmp12 -u "$username" -p "$password" -f -y >>  $log_file 2>&1
    rm -rf "${FILEMAKER_HOME}Data/Databases/devin-engine.fmp12" >>  $log_file 2>&1

    # Removing sudo entries
    color_echo "INFO" "Removing sudo entries for devin."
    if [[ "$OS" == "Darwin" ]]; then
        local temp_file=$(mktemp)
        cp -a "/etc/sudoers" "$temp_file"
        sudo sed -i '' '/^devin ALL/d' "$temp_file"
        cp -a "$temp_file" "/etc/sudoers"
    elif [[ "$OS" == "Linux" ]]; then
        sed -i '/devin/d' /etc/sudoers
    fi

    # Handle web server configuration
    if [[ "$OS" == "Darwin" ]] && grep -q "devin" "${FILEMAKER_HOME}HTTPServer/conf/extra/httpd-ssl.conf"; then
        color_echo "INFO" "Removing devin from apache."
        sed -i '' '/devin/d' "${FILEMAKER_HOME}HTTPServer/conf/extra/httpd-ssl.conf" >> $log_file 2>&1
    elif [[ "$OS" == "Linux" ]] && [ -f "${FILEMAKER_HOME}NginxServer/conf/devin.conf" ]; then
        color_echo "INFO" "Removing devin from nginx."
        sed -i '/devin/d' "${FILEMAKER_HOME}NginxServer/conf/fms_nginx.conf"
        rm "${FILEMAKER_HOME}NginxServer/conf/devin.conf"
    fi
    fmsadmin restart httpserver -u "$username" -p "$password" -f -y >>  $log_file 2>&1

    # Remove the application home directory
    rm -rf "$DEVIN_HOME"
    color_echo "SUCCESS" "Devin has been uninstalled."

    # Exit with a success status
    exit 0
}

# Function to validate the script's starting arguments
validate_arguments() {
    # Count the script's starting arguments
    local arg_count="$#"
    if [[ $arg_count -gt 0 ]]; then
        # Check if all required arguments are set
        if [[ $arg_count -ne 6 ]]; then
            color_echo "ERROR" "Invalid number of arguments. Please provide --username, --password, --type, and their respective values."
            exit 1
        fi

        # Parse arguments and assign values to variables
        while [[ $# -gt 0 ]]; do
            case "$1" in
                --username)
                    shift
                    username="$1"
                    ;;
                --password)
                    shift
                    password="$1"
                    ;;
                --type)
                    shift
                    type="$1"
                    ;;
                *)
                    color_echo "ERROR" "Invalid argument: $1"
                    exit 1
                    ;;
            esac
            shift # Move to the next key/value pair
        done

        # Check if any of the arguments are empty
        if [[ -z "$username" || -z "$password" || -z "$type" ]]; then
            color_echo "ERROR" "One or more arguments are empty. Please provide values for --username, --password, and --type."
            exit 1
        fi

        # Check if type is valid
        if [[ "$type" != "dev" && "$type" != "prod" ]]; then
            color_echo "ERROR" "Invalid value for --type. Please provide either 'dev' or 'prod'."
            exit 1
        fi
    fi
}

# Function to setup logging
setup_log() {
    local uname=$(uname -a)
    ARCH=$(uname -m)
    fmsadmin_version=$(fmsadmin -u a -p a -v < /dev/null 2>&1 | grep -o 'Version [^ ]*' | cut -d ' ' -f2 | head -n1)

    echo "############## INSTALLING DEVIN - BEGIN INSTALLER LOG FILE ##############" >> $log_file 2>&1
    echo "############## INSTALLER VERSION: $installer_version ##############" >> $log_file 2>&1
    echo -e "## Time ## \n$(date)" >> $log_file 2>&1
    echo -e "## uname -a ##\n$uname" >> $log_file 2>&1
    echo -e "## ARCH ##\n$ARCH" >> $log_file 2>&1
    echo -e "## fmsadmin version ##\n$fmsadmin_version" >> $log_file 2>&1

    if [[ "$OS" == "Darwin" ]]; then
        os_type="macOS"
        os_version=$(sw_vers -productVersion)

        echo -e "## OS Details ## \n$os_type $os_version" >> $log_file 2>&1
    elif [[ "$OS" == "Linux" ]]; then
        os_type=$(lsb_release -d | awk -F ':\t' '{print $2}')
        os_version=$(lsb_release -r | awk -F ':\t' '{print $2}')

        echo -e "## lsb_release ## \n$os_type $os_version" >> $log_file 2>&1
    fi
}

# Make sure credentials are set and valid, if not set then get them and validate
verify_credentials() {
    # Prompt for username if not set
    if [[ -z "$username" ]]; then
        color_echo "QUESTION" "Please enter your username for FileMaker Admin Console:"
        read username
    fi

    # Prompt for password if not set
    if [[ -z "$password" ]]; then
        color_echo "QUESTION" "Please enter your password for FileMaker Admin Console (will not be echoed):"
        read -s password
    fi

    # Check if it's possible to run a random fmsadmin command; if the status code == 0, then success, and we can proceed.
    fmsadmin LIST FILES -u "$username" -p "$password" > /dev/null
    exit_status=$?
    if [[ $exit_status -ne 0 ]]; then
    color_echo "ERROR" "It was not possible to validate your admin account credentials, please try again."
    exit 1
    fi
    color_echo "SUCCESS" "Credentials verified, continuing."
}

# Function to verify the version to install, dev or prod.
verify_version_to_install() {
    if [[ -z "$type" ]]; then
    color_echo "QUESTION" "Please type 0 if you are installing a Devin Development Server."
    color_echo "QUESTION" "Please type 1 if you are installing a Devin Production Server."
    read answer

    # Handle the user's input
    case $answer in
        [0]* ) 
            type=dev
            ;;
        [1]* ) 
            type=prod
            ;;
        * ) 
            color_echo "ERROR" "Please answer 0 (Development) or 1 (Production) - Exiting"
            exit 1;;
    esac
    fi
}

#On linux, ensure that systemd is installed
linux_ensure_systemd() {
    # Check if systemctl exists
    if ! which systemctl >/dev/null 2>&1; then
        color_echo "ERROR" "Systemctl (systemd) is not available on this system. You will need systemd to continue."
        read -p "Do you want to install systemd? (y/n): " install_systemd
        if [[ $install_systemd == "y" ]]; then
            color_echo "INFO" "Installing systemd."
            apt-get install -y systemd >> $log_file 2>&1
        else
            color_echo "ERROR" "Installation aborted. Please install systemd and run the script again."
            exit 1
        fi
    fi
}

ensure_jq() {
    # Check if jq is installed, if not install
    if ! which jq >/dev/null 2>&1; then
        color_echo "INFO" "Installing jq." 
        if [[ "$script_type" == "standalone" ]]; then
            if [[ "$OS" == "Darwin" ]]; then
                if [ "$ARCH" = "x86_64" ]; then
                    JQ_URL="https://github.com/jqlang/jq/releases/download/jq-1.7.1/jq-macos-amd64"
                elif [ "$ARCH" = "arm64" ]; then
                    JQ_URL="https://github.com/jqlang/jq/releases/download/jq-1.7.1/jq-macos-arm64"
                fi
                curl -k -L $JQ_URL -o /usr/local/bin/jq
                chmod +x /usr/local/bin/jq
                jq > /dev/null 2>&1
                if [[ $? -ne 2 ]]; then
                    color_echo "ERROR" "Error when running jq"
                    color_echo "INFO" "Attempting to install an older version for osx"
                    JQ_URL="https://github.com/jqlang/jq/releases/download/jq-1.6/jq-osx-amd64"
                    curl -k -L $JQ_URL -o /usr/local/bin/jq
                    chmod +x jq
                    jq > /dev/null 2>&1
                    if [[ $? -ne 2 ]]; then
                        color_echo "ERROR" "Error when running jq, please install jq manually."
                        exit 1
                    fi
                fi
            elif [[ "$OS" == "Linux" ]]; then
                apt-get install jq -y >> $log_file 2>&1
            fi
        else
            if [[ "$OS" == "Darwin" ]]; then
                cp "$script_folder_name/jq" "/usr/local/bin/jq" >> $log_file 2>&1
                chmod +x /usr/local/bin/jq >> $log_file 2>&1
                #Run jq and check the status code with $?
                jq > /dev/null 2>&1
                if [[ $? -ne 2 ]]; then
                    color_echo "ERROR" "Error when running jq"
                    color_echo "INFO" "Attempting to install an older version for osx"
                    cp "$script_folder_name/jq_osx" "/usr/local/bin/jq" >> $log_file 2>&1
                    chmod +x /usr/local/bin/jq >> $log_file 2>&1
                    jq > /dev/null 2>&1
                    if [[ $? -ne 2 ]]; then
                        color_echo "ERROR" "Error when running jq, please install jq manually."
                        exit 1
                    fi 
                fi
            elif [[ "$OS" == "Linux" ]]; then
                cp "bins/ubuntu_x64/jq" "/usr/bin/jq" >> $log_file 2>&1
                chmod +x /usr/bin/jq >> $log_file 2>&1
            fi
        fi
    fi
}


# ENsures curl is installeed
ensure_curl() {
    if ! which curl >/dev/null 2>&1; then
        # If linux run apt-get update and apt-get install curl -y
        if [[ "$OS" == "Linux" ]]; then
            color_echo "INFO" "Curl is not installed. Installing curl."
            apt-get update >> $log_file 2>&1
            apt-get install curl -y >> $log_file 2>&1
        else
            # If macOS, check if brew is installed and install curl
            if which brew >/dev/null 2>&1; then
                color_echo "INFO" "Curl is not installed. Installing curl."
                brew install curl >> $log_file 2>&1
            else 
                color_echo "ERROR" "Curl is not installed, neither is brew. Please install curl and run the script again."
                exit 1
            fi
        fi
        color_echo "ERROR" "Curl is not installed and could not be automatically installed. Please install curl and run the script again."
        exit 1
    fi
}

# Function to create and configure the devin user
create_devin_user() {
    if ! id -u devin >/dev/null 2>&1; then
        color_echo "INFO" "Creating Devin user"
        if [[ "$OS" == "Darwin" ]]; then
            # Generate a random password for macOS
            local devin_password=$(openssl rand -base64 12)
            sysadminctl -addUser devin -UID 9988 -shell /bin/bash -home /Library/Devin -password "$devin_password" >> $log_file 2>&1
            dscl . create /Users/devin IsHidden 1 >> $log_file 2>&1
        elif [[ "$OS" == "Linux" ]]; then
            useradd -u 9988 -s /bin/bash -m devin >> $log_file 2>&1
        fi
    else
        color_echo "INFO" "Devin user already exists"
    fi

    # Check if the Devin user is part of the fmsadmin group and add if not
    if [[ "$OS" == "Darwin" ]]; then
        if dscl . -read /Groups/fmsadmin GroupMembership | grep -q "devin"; then
            color_echo "INFO" "Devin user is already in the fmsadmin group"
        else
            dseditgroup -o edit -a devin -t user fmsadmin >> $log_file 2>&1
            color_echo "INFO" "Added Devin user to fmsadmin group"
        fi
    elif [[ "$OS" == "Linux" ]]; then
        if ! groups devin | grep -q '\bfmsadmin\b'; then
            color_echo "INFO" "Adding Devin user to the fmsadmin group"
            usermod -aG fmsadmin devin >> $log_file 2>&1
        else
            color_echo "INFO" "Devin user is already part of the fmsadmin group"
        fi
    fi
}

# Creates the directories
create_devin_directories() {
    #Create Devin folders
    color_echo "INFO" "Making the Devin directory."
    mkdir -p "${DEVIN_HOME}backups" >> $log_file 2>&1
    mkdir -p "${DEVIN_HOME}tmp" >> $log_file 2>&1
    mkdir -p "${DEVIN_HOME}updates" >> $log_file 2>&1
    mkdir -p "${DEVIN_HOME}logs" >> $log_file 2>&1
    mkdir -p "${DEVIN_HOME}migrations" >> $log_file 2>&1
    mkdir -p "${DEVIN_HOME}commits" >> $log_file 2>&1
    mkdir -p "${DEVIN_HOME}handler" >> $log_file 2>&1
}

# Function to check and install the FileMaker Data Migration Tool
check_and_install_dmt() {
    local exit_status=0
    local dmtlocation

    dmtlocation="${FILEMAKER_HOME}Database Server/bin/FMDataMigration"
    

    color_echo "INFO" "Checking for FileMaker Data Migration Tool at $dmtlocation."
    if [ ! -e "$dmtlocation" ]; then
        if [[ "$script_type" == "package" ]]; then
            color_echo "ERROR" "Data Migration Tool was not found"
            if [[ "$OS" == "Darwin" ]]; then
                cp "bins/mac_x64/FMDataMigration" "${FILEMAKER_HOME}Database Server/bin/FMDataMigration"
            else
                color_echo "INFO" "The data migration tool cannot be installed automatically on Linux. Please install it manually: https://community.claris.com/en/s/article/FileMaker-data-migration-tool"
                exit 1
            fi
        else
            color_echo "ERROR" "Data Migration Tool was not found in the default location."
            color_echo "INFO" "Attempting to download the Data Migration Tool."

            if [[ "$OS" == "Darwin" ]]; then
                local temp_zip="/tmp/fm_data_migration.zip"
                curl -k -o $temp_zip "https://downloads.claris.com/DEVREL/sdk/fm_data_migration_20.3.1.31_mac.zip" >> $log_file 2>&1
                unzip -o $temp_zip -d "${FILEMAKER_HOME}Database Server/bin/" >> $log_file 2>&1
                rm $temp_zip

            elif [[ "$OS" == "Linux" ]]; then
                local ubuntuversion=$(lsb_release -rs)
                curl -o "$dmtlocation" "https://devin.fm/downloads/migrationtools/FMDataMigration_U${ubuntuversion:0:2}_$fmsversion" >> $log_file 2>&1
            fi

            exit_status=$?
            if [ ! -e "$dmtlocation" ] || [[ $exit_status -ne 0 ]]; then
                color_echo "ERROR" "It was not possible to install the FileMaker Data Migration Tool, please install it manually."
                uninstall
                exit 1
            fi

            chown fmserver:fmsadmin "$dmtlocation"
            chmod 755 "$dmtlocation"
            chmod +x "$dmtlocation"
            color_echo "SUCCESS" "FileMaker Data Migration Tool was installed successfully."
        fi
    else
        color_echo "INFO" "FileMaker Data Migration Tool is already installed."
    fi
}

# Function to check FileMaker and OS version compatibility
check_version_compatibility() {


    if [[ "$OS" == "Darwin" ]]; then
        local required_macosversion="11.0"  # macOS Big Sur
        local required_fmsversion

        if [[ "$type" == "dev" ]]; then
            required_fmsversion="19.3.1"
        else
            required_fmsversion="18.0.1"
        fi

        if ! version_gt "$fmsversion" "$required_fmsversion"; then
            color_echo "ERROR" "FileMaker Server version is not compatible with Devin ($type), needs to be >= $required_fmsversion"
            exit 1
        fi

        if ! version_gt "$macosversion" "$required_macosversion"; then
            color_echo "ERROR" "macOS version is not compatible with Devin, needs to be >= macOS Big Sur (11.0)"
            exit 1
        fi

    elif [[ "$OS" == "Linux" ]]; then
        local required_linux_fmsversion="19.5.1"

        if ! version_gt "$fmsversion" "$required_linux_fmsversion"; then
            color_echo "ERROR" "FileMaker Server version is not compatible with Devin, needs to be >=19.5.1"
            exit 1
        fi
    fi
}

# Download handler to the right location
install_handler() {
    if [ "$script_type" == "standalone" ]; then
        if [ ! -f "${DEVIN_HOME}handler/devin-handler" ]; then
            color_echo "INFO" "Pulling Devin Handler Service and unzipping."
            mkdir -p ${DEVIN_HOME}handler >> $log_file 2>&1
            curl -s -o ${DEVIN_HOME}handler/handler.zip $devin_handler_download_location >> $log_file 2>&1
            unzip ${DEVIN_HOME}handler/handler.zip -d ${DEVIN_HOME}handler/ >> $log_file 2>&1
            rm ${DEVIN_HOME}handler/handler.zip >> $log_file 2>&1
            chmod +x ${DEVIN_HOME}handler/devin-handler >> $log_file 2>&1
        else
            color_echo "INFO" "${DEVIN_HOME}handler already exists"
        fi
    else
        if [ ! -f "${DEVIN_HOME}handler/devin-handler" ]; then
            color_echo "INFO" "Moving Devin Handler Service to the Devin Handler folder."
            unzip "$script_folder_name/devin-handler.zip" -d "${DEVIN_HOME}handler/"  >> $log_file 2>&1
            chmod +x ${DEVIN_HOME}handler/devin-handler >> $log_file 2>&1
        else
            color_echo "INFO" "${DEVIN_HOME}handler already exists."
        fi
    fi
}

# Download sudo utils
download_sudo_utils() {
    # Check if sudo_utils folder exists
    if [ ! -d "${DEVIN_HOME}sudo_utils" ]; then
        if [[ "$script_type" == "standalone" ]]; then
            color_echo "INFO" "Pulling utilities and unzipping."
            curl -k -s -o ${DEVIN_HOME}sudo_utils.zip "$devin_server_utils_location" >> $log_file 2>&1
        else
            cp "bins/sudo_utils.zip" "${DEVIN_HOME}sudo_utils.zip" >> $log_file 2>&1
        fi
        unzip ${DEVIN_HOME}sudo_utils.zip -d ${DEVIN_HOME}sudo_utils/ >> $log_file 2>&1
        rm ${DEVIN_HOME}sudo_utils.zip >> $log_file 2>&1
        chmod 755 ${DEVIN_HOME}sudo_utils/* >> $log_file 2>&1
    else
        color_echo "INFO" "${DEVIN_HOME}sudo_utils folder already exists, updating them"
        "${DEVIN_HOME}sudo_utils/update_sudo_utils.sh" >> $log_file 2>&1
    fi
    ${DEVIN_HOME}sudo_utils/update_sudo_utils_json.sh $username $password >> $log_file 2>&1
}

# Function to add to sudo utils
add_to_sudoers() {
    # Path for sudo_utils based on the OS
    if [ "$OS" == "Darwin" ]; then
        SUDO_UTILS_PATH="${DEVIN_HOME}sudo_utils/*"
        ADD_SUDO="sudo EDITOR='tee -a' visudo"
    elif [ "$OS" == "Linux" ]; then
        SUDO_UTILS_PATH="${DEVIN_HOME}sudo_utils/*"
        ADD_SUDO="sudo tee -a /etc/sudoers"
    else
        color_echo "ERROR" "Unsupported operating system."
        exit 1
    fi

    # Check if sudoers file exists
    if [ -f "/etc/sudoers" ]; then
        # Check if "devin" is already in the sudoers file
        if grep -q "devin" "/etc/sudoers"; then
            color_echo "INFO" "Devin is already in the sudoers list."
        else
            # Setting proper sudoers permissions
            color_echo "INFO" "Setting proper sudoers permissions"
            echo "devin ALL=(ALL) NOPASSWD: $SUDO_UTILS_PATH" | $ADD_SUDO >> $log_file 2>&1
        fi
    else
        color_echo "INFO" "Sudoers file not found."
        if [ "$OS" == "Linux" ]; then
            apt-get install -y sudo >> $log_file 2>&1
            echo "devin ALL=(ALL) NOPASSWD: $SUDO_UTILS_PATH" | $ADD_SUDO >> $log_file 2>&1
        else
            # Handle no sudoers file scenario on macOS
            color_echo "ERROR" "Sudoers file is critical for operations on macOS. Operation aborted."
            exit 1
        fi
    fi
}

# Function to add info.json
add_info_json() {
    # Check if ${DEVIN_HOME}info.json exists
    if [ ! -f "${DEVIN_HOME}info.json" ]; then
        # Set info.json
        info_version=$("${DEVIN_HOME}handler/devin-handler" -v | tr -d '\n')
        # Changing information in info.json
        color_echo "INFO" "Adding information to info.json"
        cat <<EOF > "${DEVIN_HOME}info.json"
{
    "os": {
        "name": "$os_type",
        "version": "$os_version"
    },
    "fms": {
        "version": "$fmsversion"
    },
    "devin": {
        "handler": {
            "type": "$type",
            "version": "$info_version"
        },
        "server": {
            "version": "$info_version"
        }
    }
}
EOF
    fi
}


# Function to add the devin-engine to FileMaker Server
add_devin_engine() {
    DEVIN_ENGINE_PATH="${FILEMAKER_HOME}Data/Databases/devin-engine.fmp12"

    # Install Devin developer server engine if type is dev
    if [[ "$type" == "dev" ]]; then
        if [ ! -f "$DEVIN_ENGINE_PATH" ]; then
            # Install Devin engine
            color_echo "INFO" "Installing Devin Engine"
            if [[ "$script_type" == "standalone" ]]; then
                curl -k -s -o "$DEVIN_ENGINE_PATH" "$devin_engine_download_location" >> $log_file 2>&1
            else
                cp "bins/devin-engine.fmp12" "$DEVIN_ENGINE_PATH" >> $log_file 2>&1
            fi
            chown fmserver:fmsadmin "$DEVIN_ENGINE_PATH" >> $log_file 2>&1
            chmod 755 "$DEVIN_ENGINE_PATH" >> $log_file 2>&1
        else
            color_echo "INFO" "Devin Engine already exists"
        fi

        # Validate that Devin Engine is open
        color_echo "INFO" "Validating that Devin Engine is open"
        fmsadmin -u "$username" -p "$password" LIST FILES | grep -q 'devin-engine.fmp12'
        if [[ $? -ne 0 ]]; then
            color_echo "ERROR" "The Devin Engine was not open."
            color_echo "INFO" "Attempting to open the Devin Engine."
            fmsadmin -u "$username" -p "$password" OPEN "devin-engine.fmp12" >> $log_file 2>&1
            if [[ $? -ne 0 ]]; then
                color_echo "ERROR" "It was not possible to open the Devin Engine in fmsadmin, please check the logs. - Uninstalling"
                uninstall
                exit 1
            else
                color_echo "SUCCESS" "The Devin Engine was opened successfully."
            fi
        fi
    fi
}

# function to add the prod api key to info json
add_prod_api_key() {
    # Define API key length
    API_KEY_LENGTH=15
    INFO_JSON_PATH="${DEVIN_HOME}info.json"
    # Check and create API key for production server if needed
    if [[ "$type" == "prod" ]]; then
        if ! jq -e '.devin.handler.api_key' "$INFO_JSON_PATH" > /dev/null 2>&1; then
            color_echo "INFO" "Creating Production Server secret API key"
            # Check if os is darwin
            if [[ "$OS" == "Darwin" ]]; then
                api_key=$(LC_ALL=C tr -dc 'A-Z0-9' < /dev/urandom | fold -w $API_KEY_LENGTH | head -n 1)
            else
                api_key=$(cat /dev/urandom | tr -dc 'A-Z0-9' | fold -w $API_KEY_LENGTH | head -n 1)
            fi
            jq --arg api_key "$api_key" '.devin.handler += {"api_key": $api_key}' "$INFO_JSON_PATH" > /tmp/temp.json && mv -f /tmp/temp.json "$INFO_JSON_PATH"
        fi
    fi
}

# MACOS Function to add devin as a service on with launchctl
macos_add_devin_service() {
    color_echo "INFO" "Installing Devin Handler as a service."

    # Define plist file location
    PLIST_FILE="/Library/LaunchDaemons/com.devin.handler.plist"

    # Check if Devin Handler service plist is already installed
    if [ -f "$PLIST_FILE" ]; then
        color_echo "INFO" "Devin Handler service is already installed."
    else
        # Create plist content
        cat << EOF > "$PLIST_FILE"
<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>Label</key>
    <string>com.devin.handler</string>
    <key>ProgramArguments</key>
    <array>
        <string>/Library/Devin/handler/devin-handler</string>
    </array>
    <key>RunAtLoad</key>
    <true/>
    <key>StandardOutPath</key>
    <string>/Library/Devin/logs/handler_out.log</string>
    <key>StandardErrorPath</key>
    <string>/Library/Devin/logs/handler_err.log</string>
    <key>UserName</key>
    <string>devin</string>
    <key>WorkingDirectory</key>
    <string>/Library/Devin/handler</string>
    <key>KeepAlive</key>
    <true/>
</dict>
</plist>
EOF
        if [ $? -eq 0 ]; then
            # Set appropriate permissions for plist file
            chmod 644 "$PLIST_FILE" >> $log_file 2>&1
            chown root:wheel "$PLIST_FILE" >> $log_file 2>&1

            # Load and start the launchd service using bootstrap for macOS 10.14 and later
            sudo launchctl bootstrap system "$PLIST_FILE" >> $log_file 2>&1
            if [ $? -ne 0 ]; then
                color_echo "ERROR" "Failed to load Devin Handler service."
                exit 1
            fi
        else
            color_echo "ERROR" "Failed to create plist file for Devin Handler service."
            exit 1
        fi
    fi
}

# Linux function to add devin as a service with systemd
linux_add_devin_service() {
    ## Create systemctl service for the Handler service with gunicorn
    color_echo "INFO" "Installing Devin Handler as a service."
    # Check if devin-handler service is already installed
    if systemctl list-unit-files | grep -Fqw "devin-handler"; then
        color_echo "INFO" "Devin Handler service is already installed."
    else
        read -r -d '' SERVICE_CONTENT << EOM
[Unit]
Description=Devin handler service
After=network.target

[Service]
User=devin
WorkingDirectory=/opt/Devin/handler
ExecStart=/opt/Devin/handler/devin-handler
StandardOutput=append:/opt/Devin/logs/handler_out.log
StandardError=append:/opt/Devin/logs/handler_err.log
Restart=always
KillMode=process

[Install]
WantedBy=multi-user.target
EOM

        # Define service file location
        SERVICE_FILE="/etc/systemd/system/devin-handler.service"

        # Create service file and write service content to it
        echo "${SERVICE_CONTENT}" > "${SERVICE_FILE}" 

        # Set appropriate permissions for service file
        chmod 644 "${SERVICE_FILE}" >> $log_file 2>&1

        # Reload the systemd daemon to recognize the service
        systemctl daemon-reload >> $log_file 2>&1
        systemctl daemon-reexec >> $log_file 2>&1

        #Gotta make /opt/Devin/logs owned by devin otherwise we start with an error
        chown -R devin:fmsadmin /opt/Devin/logs >> $log_file 2>&1

        # Enable the service to start on boot
        systemctl enable devin-handler >> $log_file 2>&1

        # Start it
        systemctl start devin-handler >> $log_file 2>&1
    fi
}

# Check if the handler is running
check_handler_is_running() {
    attempt=1
    while [[ $attempt -le 5 ]]; do
        color_echo "INFO" "Checking if the Devin Handler Service is running. Attempt $attempt."
        response_body=$(curl -k -s http://localhost:45712/devin/api/v1/$type/ping)
        if echo "$response_body" | grep -q "pong"; then
            color_echo "SUCCESS" "Handler Service is running successfully."
            break
        else
            if [[ $attempt -eq 5 ]]; then
                color_echo "ERROR" "Handler service did not start correctly after multiple attempts! Removing Devin and cleaning up."
                uninstall
            else
                color_echo "INFO" "Handler service not running. Retrying in 5 seconds..."
                sleep 5
            fi
        fi
        ((attempt++))
    done
}

# Mac install devin in apache
mac_install_devin_behind_apache() {
    if [[ "$type" == "prod" ]]; then
        # Define the path to your Apache configuration file
        CONFIG_FILE="${FILEMAKER_HOME}HTTPServer/conf/extra/httpd-ssl.conf"

        # Define the new line to insert
        NEW_LINE='RewriteRule ^/devin(.*)$ http://localhost:45712/devin$1 [P,L,NC]'

        # Check if the rewrite rule already exists in the config file
        if grep -q "devin" "$CONFIG_FILE"; then
            color_echo "INFO" "Rewrite rule already exists in the configuration file."
        else
            color_echo "INFO" "Adding rewrite rule to the configuration file."

            # Use sed to insert the new line before "</VirtualHost>"
            sed -i '' "/<\/VirtualHost>/i \\
            $NEW_LINE
            " "$CONFIG_FILE"


            color_echo "SUCCESS" "Rewrite rule added successfully."
        fi

        "${FILEMAKER_HOME}HTTPServer/bin/httpdctl" restart >> $log_file 2>&1
        #fmsadmin restart httpserver -u "$username" -p "$password"  -f -y >> $log_file 2>&1 #This cant be used in filemaker 19, but perhaps the above works on both

        response_body=$(curl -k https://localhost/devin/api/v1/prod/ping 2>/dev/null)
        if echo "$response_body" | grep -q "pong"; then
            color_echo "SUCCESS" "The Devin Handler Service were reachable through Apache with SSL."
        else
            color_echo "ERROR" "It was not possible to reach the Devin handler through apache, attempting to restart the HTTP server."
            #fmsadmin stop httpserver -u "$username" -p "$password" -y 
            "${FILEMAKER_HOME}HTTPServer/bin/httpdctl" stop >> $log_file 2>&1
            sleep 2
            #fmsadmin start httpserver -u "$username" -p "$password" -y 
            "${FILEMAKER_HOME}HTTPServer/bin/httpdctl" start >> $log_file 2>&1
            sleep 2
            # Retry after restart
            response_body=$(curl -k https://localhost/devin/api/v1/prod/ping 2>/dev/null)
            if echo "$response_body" | grep -q "pong"; then
                color_echo "SUCCESS" "The Devin handler was reachable through apache after restart."
            else
                color_echo "ERROR" "It was still not possible to reach the Devin handler through apache after restarting, please check that you have logged in to the admin console at least once, and chosen a certificate."
                uninstall
            fi
        fi
    fi
}

# Linux install devin behind nginx
linux_install_devin_behind_nginx() {
    # Check if prod
    if [[ "$type" == "prod" ]]; then
        # Put the handler behind nginx
        if [ ! -f "/opt/FileMaker/FileMaker Server/NginxServer/conf/devin.conf" ]; then
                color_echo "INFO" "Putting the Devin APIs behind nginx."
                cat <<EOF > "/opt/FileMaker/FileMaker Server/NginxServer/conf/devin.conf"
location ^~ /devin/ {
    proxy_set_header X-Forwarded-Proto https;
    proxy_set_header X-Forwarded-Host \$host:\$server_port;
    proxy_set_header X-Forwarded-Server \$host;
    proxy_set_header X-Forwarded-For \$proxy_add_x_forwarded_for;
    proxy_pass http://127.0.0.1:45712;
    proxy_http_version 1.1;
    proxy_set_header Connection "";
    }
EOF
        fi
        chmod 755 /opt/FileMaker/FileMaker\ Server/NginxServer/conf/devin.conf >> $log_file 2>&1
        chown fmserver:fmsadmin /opt/FileMaker/FileMaker\ Server/NginxServer/conf/devin.conf >> $log_file 2>&1

        # Check if devin entry already exists in the config file
        if ! grep -q "devin.conf" /opt/FileMaker/FileMaker\ Server/NginxServer/conf/fms_nginx.conf; then
            # Add it to main nginx config
            sed -i '/include "\/opt\/FileMaker\/FileMaker Server\/NginxServer\/conf\/fms_fac.conf";/a \
                ## devin nginx configuration\n    include "/opt/FileMaker/FileMaker Server/NginxServer/conf/devin.conf";' /opt/FileMaker/FileMaker\ Server/NginxServer/conf/fms_nginx.conf >> $log_file 2>&1
        fi

        fmsadmin restart httpserver -u "$username" -p "$password"  -f -y>>  $log_file 2>&1

        response_body=$(curl -k https://localhost/devin/api/v1/prod/ping 2>/dev/null)
        if echo "$response_body" | grep -q "pong"; then
            color_echo "SUCCESS" "The Devin Handler Service were reachable through nginx with SSL."
        else
            color_echo "ERROR" "It was not possible to reach the Devin handler through nginx, attempting to restart the HTTP server."
            fmsadmin stop httpserver -u "$username" -p "$password" -y 
            sleep 2
            fmsadmin start httpserver -u "$username" -p "$password" -y 
            sleep 2
            # Retry after restart
            response_body=$(curl -k https://localhost/devin/api/v1/prod/ping 2>/dev/null)
            if echo "$response_body" | grep -q "pong"; then
                color_echo "SUCCESS" "The Devin handler was reachable through nginx after restart."
            else
                color_echo "ERROR" "It was still not possible to reach the Devin handler through nginx after restarting, please check that you have logged in to the admin console at least once, and chosen a certificate."
                uninstall
            fi
        fi
    fi
}

# Enable odata and data api
enable_apis() {
    # Enabling the fms data api if we are on dev
    color_echo "INFO" "Enabling the Filemaker Data API."
    fmsadmin enable FMDAPI -u "$username" -p "$password" >> $log_file 2>&1
    fmsadmin start FMDAPI -u "$username" -p "$password" >> $log_file 2>&1

    color_echo "INFO" "Enabling the Filemaker OData API."
    fmsadmin enable ODATA -u "$username" -p "$password" >> $log_file 2>&1
    fmsadmin start ODATA -u "$username" -p "$password" >> $log_file 2>&1
}

# Check if handler (prod) and engine (Dev) are available from the internet.
check_available_from_interent() {
    #Getting Ip address of me
    myip=$(curl -sS ifconfig.me) >> $log_file 2>&1

    #Check if prod, that the handler is available from the internet
    if [[ "$type" == "prod" ]]; then
        color_echo "INFO" "Checking if the Devin Production Server APIs can be reached from the internet. ($myip on port 443)"
        res=$(curl -k -sS -X 'POST' 'https://portchecker.io/api/v1/query' -H 'accept: application/json' -H 'Content-Type: application/json' -d '{"host": "'$myip'","ports": ["443"]}')  >> $log_file 2>&1
        status=$(echo $res | jq -r '.check[0].status')

        if [[ "$status" != "true" ]]; then
            color_echo "ERROR" "It was not possible to ensure that port 443 is reachable from the internet. If you need Devin to reach the production server over the internet, then this is a problem that you will need to investigate."
        else
            color_echo "SUCCESS" "It was possible to reach the Devin Production Server APIs from the internet."
        fi
    fi

    #Check if dev, that the engine is available from the internet
    if [[ "$type" == "dev" ]]; then
        color_echo "INFO" "Checking if the Devin Engine can be reached from the internet. ($myip on port 5003)"
        res=$(curl -k -sS -X 'POST' 'https://portchecker.io/api/v1/query' -H 'accept: application/json' -H 'Content-Type: application/json' -d '{"host": "'$myip'","ports": [5003]}')  >> $log_file 2>&1
        status=$(echo $res | jq -r '.check[0].status')

        if [[ "$status" != "true" ]]; then
            color_echo "ERROR" "It was not possible to ensure that port 5003 is reachable from the internet. If you need Devin to be accessible over internet, then you will need to investigate this."
        else
            color_echo "SUCCESS" "It was possible to reach the Devin Engine from the internet."
        fi
    fi
}

#Install filemaker script schedule
install_script_schedule_via_admin_api() {
    #Check if dev, then enable script schedule in fmi admin api
    if [[ "$type" == "dev" ]]; then
        success=false
        color_echo "INFO" "Enabling the Filemaker Script Schedule."
        # Make the request and extract the token in one line
        auth=$(echo -n "$username:$password" | base64 | tr -d '\n')
        token=$(curl -k -s -X POST "https://localhost/fmi/admin/api/v2/user/auth" -H "Content-Type: application/json" -H "Authorization: Basic $auth" | jq -r '.response.token')

        if [ "$token" != "null" ] || [ -n "$token" ]; then
            # Check if the schedule somehow is there:
            existingSchedules=$(curl -k -s  -X GET "https://localhost/fmi/admin/api/v2/schedules" \
            -H "Content-Type: application/json" \
            -H "Authorization: Bearer $token")


            scheduleId=$(echo "$existingSchedules" | jq -r '.response.schedules[] | select(.filemakerScriptType.fmScriptAccount == "DevinEngineScheduleRunner").id')
            if [ -n "$scheduleId" ]; then
                color_echo "INFO" "The script schedule already exists, removing it."
                curl -k -s  -X DELETE "https://localhost/fmi/admin/api/v2/schedules/$scheduleId" \
                -H "Content-Type: application/json" \
                -H "Authorization: Bearer $token" >> $log_file 2>&1
            fi

            jsonPayload='{
            "name": "Devin Engine Assistant",
            "filemakerScriptType": {
                "resource": "devin-engine.fmp12",
                "fmScriptName": "DevinEngine1MinRefresh",
                "fmScriptPassword": "12345678",
                "timeout": 1,
                "autoAbort": true,
                "fmScriptAccount": "DevinEngineScheduleRunner"
            },
            "enabled": true,
            "everyndaysType": {
                "dailyDays": 1,
                "repeatTask": {
                "endTime": "23:59:00",
                "repeatFrequency": 1,
                "repeatInterval": "MINUTES"
                },
                "startTimeStamp": "2022-01-01T00:00:00"
            }
            }'

            # Send the POST request and save the output to a variable
            response=$(curl  -k -s  -X POST "https://localhost/fmi/admin/api/v2/schedules/filemakerscript" \
            -H "Content-Type: application/json" \
            -H "Authorization: Bearer $token" \
            -d "$jsonPayload")
            echo $response >> $log_file 2>&1
            messageCode=$(echo "$response" | jq -r '.messages[0].code')
            # Check if the message code is not "0"
            if [ "$messageCode" != "0" ]; then
                color_echo "ERROR" "Something went wrong when starting the script schedule. Code: $messageCode, make sure that the script schedule does not already exist."
            else
                color_echo "SUCCESS" "Successfully authenticated to the admin api and created the script schedule"
                success=true
            fi
        else
            color_echo "ERROR" "Failed to authenticate to the admin api and start the script schedule, reverting installation"
        fi
        curl -k -s  -X DELETE "https://localhost/fmi/admin/api/v2/user/auth/$token" >> $log_file 2>&1
        if [[ "$success" == "false" ]]; then
            uninstall
        fi
    fi
}

# Set permissions on the devin folder and files
set_permissions() {
    color_echo "INFO" "Setting permissions on Devin folder."

    # Set directory permissions
    find $DEVIN_HOME -type d -not -path "*sudo_utils*" -exec chown devin:fmsadmin {} \; -exec chmod 755 {} \; >> $log_file 2>&1

    # Set file permissions
    find $DEVIN_HOME -type f -not -path "*sudo_utils*"  -exec chown devin:fmsadmin {} \; -exec chmod 644 {} \; >> $log_file 2>&1

    chmod +x "${DEVIN_HOME}handler/devin-handler" >> $log_file 2>&1
}


# Echo prod api key to user
echo_prod_api_key() {
    #Echo prod api key
    if [[ "$type" == "prod" ]]; then
    color_echo "SUCCESS" "####################################################################################"
    color_echo "SUCCESS" "###       Your api key for this Production Server is:     $api_key       ####"
    color_echo "SUCCESS" "####################################################################################"
    color_echo "SUCCESS" "Please save the key, as you will need it when adding this server to your Devin App."
    color_echo "SUCCESS" "If you forget it, you can find it in /Library/Devin/info.json"
    fi
}

# Add the uninstaller script 
install_uninstaller() {
    if [[ $script_type == "standalone" ]]; then
        color_echo "INFO" "Downloading the uninstaller script."
        curl -k -s -o "${DEVIN_HOME}uninstall_devin.sh" "$devin_uninstaller_location" >> $log_file 2>&1
        chmod 755 "${DEVIN_HOME}uninstall_devin.sh" >> $log_file 2>&1
    else
        cp "bins/uninstall_devin.sh" "${DEVIN_HOME}uninstall_devin.sh" >> $log_file 2>&1
        chmod 755 "${DEVIN_HOME}uninstall_devin.sh" >> $log_file 2>&1
    fi
}






######### INSTALLER CALL FUNCTIIONS #########

log_file="devin_installer_$(date +%Y%m%d_%H%M%S).log"
devin_engine_download_location="https://download.devin.fm/downloads/server/latest/devin-engine.fmp12"
devin_server_utils_location="https://download.devin.fm/downloads/server/latest/sudo_utils.zip"
devin_uninstaller_location="https://download.devin.fm/downloads/server/latest/uninstall_devin.sh"
OS=$(uname)

# Set some variables depending on the OS
case $OS in
    "Darwin")
        #Check if arm or x64
        if [[ $(uname -m) == "arm64" ]]; then
            devin_handler_download_location="https://download.devin.fm/downloads/server/latest/mac_arm_handler.zip"
        elif [[ $(uname -m) == "x86_64" ]]; then
            devin_handler_download_location="https://download.devin.fm/downloads/server/latest/mac_x64_handler.zip"
        fi
        DEVIN_HOME="/Library/Devin/"
        FILEMAKER_HOME="/Library/FileMaker Server/"
        macosversion=$(sw_vers -productVersion)
        fmsversion=$(fmsadmin -u a -p a -v < /dev/null | grep -o 'Version [^ ]*' | cut -d ' ' -f2)
        ;;
    "Linux")
        # Check if arm or x64
        if [[ $(uname -m) == "aarch64" ]]; then
            devin_handler_download_location="https://download.devin.fm/downloads/server/latest/ubuntu_arm_handler.zip"
        elif [[ $(uname -m) == "x86_64" ]]; then
            devin_handler_download_location="https://download.devin.fm/downloads/server/latest/ubuntu_x64_handler.zip"
        fi
        DEVIN_HOME="/opt/Devin/"
        FILEMAKER_HOME="/opt/FileMaker/FileMaker Server/"
        fmsversion=$(fmsadmin -u a -p a -v < /dev/null | grep -oP 'fmsadmin: Version \K\S+')
        ubuntuversion=$(lsb_release -d | cut -d ' ' -f2)
        ;;
    *)
        color_echo "ERROR" "Unsupported OS."
        exit 1
        ;;
esac



# From here call all the functions from top to botom, but check if for the os specific ones for the right os
color_echo "INFO" "Devin installer starting. Please consult the '$log_file' file if you encounter errors."

# Check if the script is run as root
ensure_root
# Check if fms is installed
ensure_fms_installed
# Validate the arguments if present
validate_arguments "$@"
# Setup log
setup_log
# Verify user credentials
verify_credentials
# Check if its a package or standalone
check_packaged_or_standalone
# Verify the version to install
verify_version_to_install
# If on Linux, ensure systemd is available
if [[ "$OS" == "Linux"* ]]; then
    linux_ensure_systemd
fi
# Ensure curl is installed if the script type is standalone
if [[ "$script_type" == "standalone" ]]; then
    ensure_curl
fi
# Ensure jq is installed or gets installed
ensure_jq
# Create a user for devin
create_devin_user
# Create necessary directories for devin
create_devin_directories
# Check and install DMT if necessary
check_and_install_dmt
# Check compatibility of the version
check_version_compatibility
# Download the handler
install_handler
# Download sudo utilities
download_sudo_utils
# Add the user to sudoers
add_to_sudoers
# Add information to JSON file
add_info_json
# Add devin engine
add_devin_engine
# Add production API key
add_prod_api_key
# If on MacOS, add devin service
if [[ "$OS" == "Darwin"* ]]; then
    macos_add_devin_service
fi
# If on Linux, add devin service
if [[ "$OS" == "Linux"* ]]; then
    linux_add_devin_service
fi
# Check if the handler is running
check_handler_is_running
# If on MacOS, install devin behind Apache
if [[ "$OS" == "Darwin"* ]]; then
    mac_install_devin_behind_apache
fi
# If on Linux, install devin behind Nginx
if [[ "$OS" == "Linux"* ]]; then
    linux_install_devin_behind_nginx
fi
# Enable APIs
enable_apis
# Check if the service is available from the internet
check_available_from_interent
# Schedule the install script via admin API
install_script_schedule_via_admin_api
# Install uninstaller
install_uninstaller
# Set necessary permissions
set_permissions
# Echo the production API key
echo_prod_api_key